import datetime
import operator
from functools import reduce

import phonenumbers
import pytz
from dateutil.parser import parse as parse_date

from tztrout.data import ZIP_DATA, TroutData
from tztrout.data_exceptions import data_exceptions

td = TroutData()


def _dedupe_preserve_ord(lst):
    """Dedupe a list and preserve the order of its items."""
    seen = set()
    return [x for x in lst if x not in seen and not seen.add(x)]


def tz_ids_for_tz_name(tz_name):
    """Get the TZ identifiers that are currently in a specific time zone, e.g.

    >>> tztrout.tz_ids_for_tz_name('PDT')  # ran during DST
    [
        u'America/Dawson',
        u'America/Los_Angeles',
        u'America/Santa_Isabel',
        u'America/Tijuana',
        u'America/Vancouver',
        u'America/Whitehorse',
        u'Canada/Pacific',
        u'US/Pacific'
    ]
    >>> tztrout.tz_ids_for_tz_name('PDT')  # ran outside of the DST period
    []
    """
    ids = td.tz_name_to_tz_ids.get(tz_name)

    # if the tz_name is just an alias, don't perform the fine-grained filtering
    if tz_name in td.aliases:
        return ids

    valid_ids = []
    if ids:
        # only get the tz ids that match the tz name currently
        for id in ids:
            tz = pytz.timezone(id)
            try:
                if tz.tzname(datetime.datetime.utcnow()) == tz_name:
                    valid_ids.append(id)
            except (pytz.NonExistentTimeError, pytz.AmbiguousTimeError):
                pass
    return valid_ids


def tz_ids_for_phone(phone, country='US'):
    """Get the TZ identifiers that a phone number might be related to, e.g.

    >>> tztrout.tz_ids_for_phone('+16503334444')
    [u'America/Los_Angeles']
    >>> tztrout.tz_ids_for_phone('+49 (0)711 400 40990')
    [u'Europe/Berlin', u'Europe/Busingen']
    """
    from phonenumbers.geocoder import description_for_number

    try:
        phone = phonenumbers.parse(phone, country)
    except Exception:
        pass
    else:
        country_iso = phonenumbers.region_code_for_number(phone)
        if not country_iso:
            country_iso = phonenumbers.region_code_for_country_code(
                phone.country_code
            )

        if country_iso == 'US':
            # check if we have a specific exception for a given area code first
            exception_key = 'areacode:%s' % str(phone.national_number)[:3]
            if exception_key in data_exceptions:
                return data_exceptions[exception_key]['include']

            state = city = None
            area = description_for_number(phone, 'en').split(',')
            if len(area) == 2:
                city = area[0].strip()
                state = area[1].strip()
            elif len(area) == 1 and area[0]:
                state = area[0].lower().strip()
                state = td.normalized_states['US'].get(state, None)

            return tz_ids_for_address(country_iso, state=state, city=city)

        elif country_iso == 'CA':
            area_code = str(phone.national_number)[:3]
            state = td.ca_area_code_to_state.get(area_code)
            return td.ca_state_to_tz_ids.get(state)

        elif country_iso == 'AU':
            area_code = str(phone.national_number)[:2]
            state = td.au_area_code_to_state.get(area_code)

            # Some Australian number prefixes (e.g. 04) are country-wide - fall
            # back to all the AU tz ids
            if state:
                return td.au_state_to_tz_ids.get(state)
            return pytz.country_timezones.get(country_iso)

        elif country_iso:
            return pytz.country_timezones.get(country_iso)

    return []


def tz_ids_for_address(country, state=None, city=None, zipcode=None, **kwargs):
    """Get the TZ identifiers for a given address, e.g.:

    >>> tztrout.tz_ids_for_address('US', state='CA', city='Palo Alto')
    [u'America/Los_Angeles']
    >>> tztrout.tz_ids_for_address('PL')
    [u'Europe/Warsaw']
    >>> tztrout.tz_ids_for_address('CN')
    [
        u'Asia/Shanghai',
        u'Asia/Harbin',
        u'Asia/Chongqing',
        u'Asia/Urumqi',
        u'Asia/Kashgar'
    ]
    """
    if country == 'US':
        if zipcode:
            if isinstance(zipcode, int):
                zipcode = str(zipcode)

            # If an extended zipcode in a form of XXXXX-XXXX is provided,
            # only use the first part
            zipcode = zipcode.split('-')[0]

            return list(td.us_zip_to_tz_ids.get(zipcode, []))
        elif state or city:
            if city and 'city:%s' % city.lower() in data_exceptions:
                return data_exceptions['city:%s' % city.lower()]['include']
            if state and len(state) != 2:
                state = td.normalized_states['US'].get(state.lower(), state)
            code = ZIP_DATA.find_zip(city=city, state=state)
            if code:
                return list(td.us_zip_to_tz_ids.get(code, []))
            elif city and state:
                # Couldn't find by city+state, try ignoring city
                code = ZIP_DATA.find_zip(state=state)
                if code:
                    return list(td.us_zip_to_tz_ids.get(code, []))
    elif country == 'CA' and state:
        if len(state) != 2:
            state = td.normalized_states['CA'].get(state.lower(), state)
        return td.ca_state_to_tz_ids.get(state)
    elif country == 'AU' and state:
        if len(state) != 2:
            state = td.normalized_states['AU'].get(state.lower(), state)
        return td.au_state_to_tz_ids.get(state)

    return pytz.country_timezones.get(country)


def tz_ids_for_offset(offset_in_minutes):
    """Get the TZ identifiers for a given UTC offset (in minutes), e.g.

    >>> tztrout.tz_ids_for_offset(-7 * 60)  # during DST
    [
        u'America/Creston',
        u'America/Dawson',
        u'America/Dawson_Creek',
        u'America/Hermosillo',
        u'America/Los_Angeles',
        u'America/Phoenix',
        u'America/Santa_Isabel',
        u'America/Tijuana',
        u'America/Vancouver',
        u'America/Whitehorse',
        u'Canada/Pacific',
        u'US/Arizona',
        u'US/Pacific'
    ]
    """
    ids = td.offset_to_tz_ids.get(offset_in_minutes)
    valid_ids = []
    if ids:
        # only get the tz ids that match the tz offset currently
        for id in ids:
            tz = pytz.timezone(id)
            try:
                off = tz.utcoffset(datetime.datetime.utcnow()).total_seconds()
                if off / 60 == offset_in_minutes:
                    valid_ids.append(id)
            except (pytz.NonExistentTimeError, pytz.AmbiguousTimeError):
                pass
    return valid_ids


def local_time_for_phone(phone, country='US'):
    """Get the local time for a given phone number, e.g.

    >>> datetime.datetime.utcnow()
    datetime.datetime(2013, 9, 17, 19, 44, 0, 966696)
    >>> tztrout.local_time_for_phone('+16503334444')
    datetime.datetime(2013, 9, 17, 12, 44, 0, 966696, tzinfo=<DstTzInfo 'America/Los_Angeles' PDT-1 day, 17:00:00 DST>)
    """
    ids = tz_ids_for_phone(phone, country)
    if ids:
        return pytz.timezone(ids[0]).fromutc(datetime.datetime.utcnow())


def local_time_for_address(
    country, state=None, city=None, zipcode=None, **kwargs
):
    """Get the local time for a given address, e.g.

    >>> datetime.datetime.utcnow()
    datetime.datetime(2013, 9, 17, 19, 44, 0, 966696)
    >>> tztrout.local_time_for_address('US', state='California')
    datetime.datetime(2013, 9, 17, 12, 44, 0, 966696, tzinfo=<DstTzInfo 'America/Los_Angeles' PDT-1 day, 17:00:00 DST>)
    """
    ids = tz_ids_for_address(country, state, city, zipcode, **kwargs)
    if ids:
        return pytz.timezone(ids[0]).fromutc(datetime.datetime.utcnow())


def offset_ranges_for_local_time(local_start, local_end):
    """Return a list of UTC offset ranges matching a given local time range.

    The returned ranges are expressed in minutes.

    `local_start` and `local_end` can be instances of `datetime.time`,
    integers, (minutes from midnight) or strings (e.g. '10:00', '20:00',
    '8pm', etc.)

    For example, if we want to find offset ranges for the timezones where the
    local time is currently between 9am and 5pm, and it's 8pm UTC at the
    moment, then the expected return value is:
    * From -660 to -180. That's because:
        * 20 - 660 / 60 == 9
        * 20 - 180 / 60 == 17
    * From 780 to 840. That's because:
        * 20 + 780 / 60 % 24 == 9
        * 20 + 840 / 60 % 24 == 10 (it's not 5pm, but 840/60 is 14 and UTC+14
          is the furthest possible timezone offset).

    >>> tztrout.offset_ranges_for_local_time(datetime.time(9), datetime.time(17))  # ran at 8pm UTC
    [[-660, -180], [780, 840]]
    """
    if not isinstance(local_start, (datetime.time, int, str)):
        raise ValueError(
            'local_start is not an instance of datetime.time or int or str'
        )
    if not isinstance(local_end, (datetime.time, int, str)):
        raise ValueError(
            'local_end is not an instance of datetime.time or int or str'
        )

    # Convert to `datetime.time` if `local_start` or `local_end` are strings.
    local_start = (
        parse_date(local_start).time()
        if isinstance(local_start, str)
        else local_start
    )
    local_end = (
        parse_date(local_end).time()
        if isinstance(local_end, str)
        else local_end
    )

    # Convert to ints (minutes).
    def to_minutes(t):
        return t.hour * 60 + t.minute

    local_start = (
        local_start
        if isinstance(local_start, int)
        else to_minutes(local_start)
    )
    local_end = (
        local_end if isinstance(local_end, int) else to_minutes(local_end)
    )

    # Get current UTC time.
    current_time = to_minutes(datetime.datetime.utcnow().time())

    # Tweak for ranges that pass midnight (e.g. (5pm, 9am)).
    if local_end < local_start:
        local_end += 24 * 60

    # Calculate UTC offsets.
    offset_ranges = [
        [local_start - current_time, local_end - current_time],
        [
            24 * 60 - current_time + local_start,
            24 * 60 - current_time + local_end,
        ],
        [
            -24 * 60 - current_time + local_start,
            -24 * 60 - current_time + local_end,
        ],
    ]

    # Cap the offsets at UTC-14:00 and UTC+14:00 (lowest/highest possible
    # offset).
    def capped(t):
        return 14 * 60 if t > 14 * 60 else -14 * 60 if t < -14 * 60 else t

    for i, range in enumerate(offset_ranges):
        offset_ranges[i][0] = capped(range[0])
        offset_ranges[i][1] = capped(range[1])

    # Discard the irrelevant ranges (i.e. where start == end).
    offset_ranges = [range for range in offset_ranges if range[0] != range[1]]
    return offset_ranges


def tz_ids_for_offset_range(offset_start, offset_end):
    """Return all the time zone identifiers which offsets are within the
    (offset_start, offset_end) range. The arguments should be integers
    (UTC offsets in minutes).

    >>> tztrout.tz_ids_for_offset_range(-7*60, -6 *60)
    [
        u'America/Belize', u'America/Boise', u'America/Cambridge_Bay',
        u'America/Chihuahua', u'America/Costa_Rica', u'America/Denver',
        u'America/Edmonton', u'America/El_Salvador', u'America/Guatemala',
        u'America/Inuvik', u'America/Managua', u'America/Mazatlan',
        u'America/Ojinaga', u'America/Regina', u'America/Shiprock',
        u'America/Swift_Current', u'America/Tegucigalpa',
        u'America/Yellowknife', u'Canada/Mountain', u'Pacific/Galapagos',
        u'US/Mountain', u'America/Creston', u'America/Dawson',
        u'America/Dawson_Creek', u'America/Hermosillo', u'America/Los_Angeles',
        u'America/Phoenix', u'America/Santa_Isabel', u'America/Tijuana',
        u'America/Vancouver', u'America/Whitehorse', u'Canada/Pacific',
        u'US/Arizona', u'US/Pacific'
    ]
    """
    offsets = [
        int(o)
        for o in td.offset_to_tz_ids.keys()
        if (int(o) >= offset_start and int(o) <= offset_end)
    ]
    ids = [tz_ids_for_offset(o) for o in offsets]
    if ids:
        ids = reduce(operator.add, ids)  # flatten the list of lists
    return ids


def non_dst_offsets_for_phone(phone):
    """Return the non-DST offsets (in minutes) for a given phone, e.g.

    >>> tztrout.non_dst_offsets_for_phone('+1 650 248 6188')
    [-480]
    """
    ids = tz_ids_for_phone(phone)
    if ids:
        offsets = [
            td._get_latest_non_dst_offset(pytz.timezone(id)) for id in ids
        ]
        return _dedupe_preserve_ord(
            [int(o.total_seconds() / 60) for o in offsets if o]
        )


def non_dst_offsets_for_address(
    country, state=None, city=None, zipcode=None, **kwargs
):
    """Return the non-DST offsets (in minutes) for a given address, e.g.

    >>> tztrout.non_dst_offsets_for_address('US', state='CA')
    [-480]
    """
    ids = tz_ids_for_address(
        country, state=state, city=city, zipcode=zipcode, **kwargs
    )
    if ids:
        offsets = [
            td._get_latest_non_dst_offset(pytz.timezone(id)) for id in ids
        ]
        return _dedupe_preserve_ord(
            [int(o.total_seconds() / 60) for o in offsets if o]
        )



"""Put the exceptions for the auto-generated data here. The format should be:

'type:value': {
    'include': [list, of, tz, ids],
    'exclude': [list, of, tz, ids]
}

Allowed types are: ('zip', 'state', 'all', 'city', 'areacode').

Note that exceptions for cities and area codes have a higher priority than any
other timezone rule. E.g. if you do 'city:new york': { 'include': ['America/New_York'] },
that's *the only* tz id to be returned for tztrout.tz_ids_for_address('US', state='NY', city='New York').
"""

data_exceptions = {
    'all': {
        'exclude': [
            'America/Indiana/Tell_City',
            'America/Indiana/Knox',
            'America/North_Dakota/Beulah',
            'America/Indiana/Winamac',
            'America/Indiana/Petersburg',
            'America/Indiana/Vincennes',
        ]
    },
    'state:NY': {
        'exclude': [
            'America/Indiana/Winamac',
            'America/Indiana/Petersburg',
            'America/Indiana/Vincennes',
        ]
    },
    'state:PA': {
        'exclude': [
            'America/Indiana/Winamac',
            'America/Indiana/Petersburg',
            'America/Indiana/Vincennes',
        ]
    },
    'state:IN': {'include': ['America/Indiana/Indianapolis']},
    'city:el paso': {'include': ['America/Denver']},
    'areacode:915': {'include': ['America/Denver']},
    'areacode:901': {'include': ['America/Chicago']},
    'areacode:202': {'include': ['America/New_York']},
    'areacode:502': {'include': ['America/New_York']},  # eastern Kentucky
    'areacode:859': {'include': ['America/New_York']},  # eastern Kentucky
}



import datetime
import json
import os
from collections import defaultdict, namedtuple
from sys import stdout

import pytz

from tztrout.data_exceptions import data_exceptions

# paths to the data files
basepath = os.path.dirname(os.path.abspath(__file__))
US_ZIPS_TO_TZ_IDS_MAP_PATH = os.path.join(
    basepath, 'data/us_zips_to_tz_ids.json'
)
TZ_NAME_TO_TZ_IDS_MAP_PATH = os.path.join(
    basepath, 'data/tz_name_to_tz_ids.json'
)
OFFSET_TO_TZ_IDS_MAP_PATH = os.path.join(
    basepath, 'data/offset_to_tz_ids.json'
)
STATES_PATH = os.path.join(basepath, 'data/normalized_states.json')
ALIASES_PATH = os.path.join(basepath, 'data/tz_name_to_tz_name_aliases.json')
CA_STATE_TO_TZ_IDS_MAP_PATH = os.path.join(
    basepath, 'data/ca_state_to_tz_ids.json'
)
CA_AREA_CODE_TO_STATE_MAP_PATH = os.path.join(
    basepath, 'data/ca_area_code_to_state.json'
)
AU_STATE_TO_TZ_IDS_MAP_PATH = os.path.join(
    basepath, 'data/au_state_to_tz_ids.json'
)
AU_AREA_CODE_TO_STATE_MAP_PATH = os.path.join(
    basepath, 'data/au_area_code_to_state.json'
)

# Australian TZ names are resolved as EST, WST, etc., which is ok for local
# usage, but conflicts with the more common US TZ names if we're thinking
# globally. We should use AWST, ACST, AEST, etc. instead.
AU_MAP = {
    'WST': 'AWST',
    'CST': 'ACST',
    'EST': 'AEST',
    'CWST': 'AWST',
    'LHST': 'AEST',
    'WDT': 'AWDT',
    'CDT': 'ACDT',
    'EDT': 'AEDT',
    'CWDT': 'AWDT',
    'LHDT': 'AEDT',
}

# Asia/Manila is resolved as PST (Philippine Standard Time), which is ok for
# local usage, but conflicts with the more common US PST (Pacific Standard
# Time). We use PHT instead.
ASIA_MAP = {
    'PST': 'PHT',
}

# The path of our US Zipcode data generated from Geonames.org
US_ZIPCODE_DATA_PATH = os.path.join(basepath, 'data/us_zipcode_data.json')

# The namedtuple structure that represents a US Zipcode
Zipcode = namedtuple(
    'Zipcode', ['zip', 'city', 'state', 'latitude', 'longitude']
)


def deduplicator():
    """Deduplicate strings in memory using a canonical mapping.

    Works similarly to intern() but supports unicode as well.

    >>> a = chr(1) + chr(1)  # Create two strings
    >>> b = chr(1) + chr(1)

    >>> a == b  # These have the same contents
    True
    >>> id(a) == id(b)  # But they're different objects
    False

    >>> deduplicate = deduplicator()  # Deduplicating them returns the same object
    >>> id(deduplicate(a)) == id(deduplicate(b))
    True
    """
    memo = {}

    def deduplicate(item):
        if item in memo:
            return memo[item]
        memo[item] = item
        return item

    return deduplicate


def generate_us_zipcode_namedtuples():
    """Generate a list of namedtuples for US zipcodes from the list of
    dictionaries in data/us_zipcode_data.json
    """
    if not os.path.exists(US_ZIPCODE_DATA_PATH):
        raise ValueError(
            'US Zipcode data missing. Run regenerate_data.py first'
        )
    with open(US_ZIPCODE_DATA_PATH) as file:
        data = json.load(file)
    for zip in data:
        yield Zipcode(**zip)


class InMemoryZipData:
    """In-memory copy of Zipcode mappings by state, city, and state and city.

    This enables 20x speedup for city/state zip code lookup
    at a cost of more restricted / use-case-specific API,
    extra startup time of about 0.25-0.5s (depending on CPU),
    and holding the dataset in process memory (4-8MB).
    """

    def __init__(self):
        self.by_state = {}
        self.by_city = {}
        self.by_state_city = {}
        deduplicate = deduplicator()

        for zip in generate_us_zipcode_namedtuples():
            code = deduplicate(zip.zip)
            state = deduplicate(zip.state)
            city = deduplicate(zip.city)
            if state not in self.by_state:
                self.by_state[state] = code
            if city not in self.by_city:
                self.by_city[city] = code
            if (state, city) not in self.by_state_city:
                self.by_state_city[(state, city)] = code

    def find_zip(self, city=None, state=None):
        if city and state:
            return self.by_state_city.get((state, city))
        elif city:
            return self.by_city.get(city)
        elif state:
            return self.by_state.get(state)

        raise ValueError('Specify either city or state')


ZIP_DATA = InMemoryZipData()


class TroutData:
    """Helper class that generates the JSON data used by TZTrout"""

    # We don't care about the historic data - we just want to know the recent
    # state of time zones, zip codes, etc. RECENT_YEARS_START describes how
    # far back we should go to check for DST changes, timezone names, etc.
    RECENT_YEARS_START = 2005

    # Sometimes we need to go back a few steps to figure out DSTs, timezone
    # names, etc. This is a kwargs dict that is passed to datetime.timedelta
    # to determine the size of a single step
    TD_STEP = {'days': 40}

    def __init__(self):
        # load the data files, if they exist
        self._load_us_zipcode_data(
            'us_zip_to_tz_ids', US_ZIPS_TO_TZ_IDS_MAP_PATH
        )
        self._load_data('tz_name_to_tz_ids', TZ_NAME_TO_TZ_IDS_MAP_PATH)
        self._load_data('offset_to_tz_ids', OFFSET_TO_TZ_IDS_MAP_PATH)
        self._load_data('normalized_states', STATES_PATH)
        self._load_data('tz_name_aliases', ALIASES_PATH)
        self._load_data('ca_state_to_tz_ids', CA_STATE_TO_TZ_IDS_MAP_PATH)
        self._load_data(
            'ca_area_code_to_state', CA_AREA_CODE_TO_STATE_MAP_PATH
        )
        self._load_data('au_state_to_tz_ids', AU_STATE_TO_TZ_IDS_MAP_PATH)
        self._load_data(
            'au_area_code_to_state', AU_AREA_CODE_TO_STATE_MAP_PATH
        )

        # convert string offsets into integers
        self.offset_to_tz_ids = {
            int(k): v for k, v in self.offset_to_tz_ids.items()
        }

        # flatten the values of all tz name aliases (needed to identify which
        # filters don't need to be exact in tztrout.tz_ids_for_tz_name
        self.aliases = {
            alias for v in self.tz_name_aliases.values() for alias in v
        }

    def _load_data(self, name, path):
        """Helper method to load a data file"""
        with open(path) as file:
            data = json.loads(file.read())
            setattr(self, name, data)

    def _load_us_zipcode_data(self, name, path):
        dedupe = deduplicator()
        with open(path) as file:
            data = json.load(file)
            setattr(
                self,
                name,
                {
                    dedupe(k): dedupe(tuple(dedupe(tz) for tz in v))
                    for k, v in data.items()
                },
            )

    def _get_latest_non_dst_offset(self, tz):
        """Get the UTC offset for a given time zone identifier. Ignore the
        DST offsets.
        """
        dt = datetime.datetime.utcnow()
        while dt.year > self.RECENT_YEARS_START:
            try:
                dst = tz.dst(dt).total_seconds()
                if not dst:
                    return tz.utcoffset(dt)
            except (pytz.NonExistentTimeError, pytz.AmbiguousTimeError):
                pass
            dt -= datetime.timedelta(**self.TD_STEP)

    def _get_latest_offsets(self, tz):
        """Get all the UTC offsets (in minutes) that a given time zone
        experienced in the recent years.
        """
        dt = datetime.datetime.utcnow()
        offsets = []
        while dt.year > self.RECENT_YEARS_START:
            try:
                offset = int(tz.utcoffset(dt).total_seconds() / 60)
                if offset not in offsets:
                    offsets.append(offset)
            except (pytz.NonExistentTimeError, pytz.AmbiguousTimeError):
                pass
            dt -= datetime.timedelta(**self.TD_STEP)
        return offsets

    def _get_latest_tz_names(self, tz):
        """Get the recent time zone names for a given time zone identifier."""
        dt = datetime.datetime.utcnow()
        tz_names = []
        while dt.year > self.RECENT_YEARS_START:
            try:
                tz_name = tz.tzname(dt)

                # Ignore TZ names that are really UTC offsets like "+01".
                is_offset = tz_name.startswith(('+', '-'))
                if not is_offset and tz_name not in tz_names:
                    tz_names.append(tz_name)
            except (pytz.NonExistentTimeError, pytz.AmbiguousTimeError):
                pass
            dt -= datetime.timedelta(**self.TD_STEP)
        return tz_names

    def generate_zip_to_tz_id_map(self):
        """Generate the map of US zipcodes to time zone identifiers. The
        method finds all the possible time zone identifiers for each zipcode
        using the latitude and longitude of each zipcode to search
        TimezoneFinder.
        """
        try:
            from timezonefinder import TimezoneFinder
        except ImportError:
            raise ImportError(
                'Dependency "timezonefinder" is missing. '
                'Install with "tz-trout[dev]" in order to generate timezone data'
            )

        tf = TimezoneFinder()

        def _get_tz_identifiers_for_us_zipcode(zipcode):
            """Get all the possible identifiers for a given US zip code."""
            # Find the TZ identifier for a Zipcode given its latitude and longitude
            # using the TimezoneFinder library
            tz_id = tf.timezone_at(lng=zipcode.longitude, lat=zipcode.latitude)
            return [tz_id]

        cnt = 0
        tz_ids_to_zips = defaultdict(list)
        for zip in generate_us_zipcode_namedtuples():
            ids = tuple(_get_tz_identifiers_for_us_zipcode(zip))
            cnt += 1
            if cnt % 100 == 99:
                stdout.write(f'\r{cnt} zipcodes mapped to tz_ids')
                stdout.flush()

            # apply the data exceptions
            exceptions = (
                data_exceptions.get('zip:' + zip.zip)
                or data_exceptions.get('state:' + zip.state)
                or {}
            )
            exceptions['include'] = (
                exceptions.get('include', [])
                + data_exceptions['all'].get('include', [])
                if 'all' in data_exceptions
                else []
            )
            exceptions['exclude'] = (
                exceptions.get('exclude', [])
                + data_exceptions['all'].get('exclude', [])
                if 'all' in data_exceptions
                else []
            )
            if exceptions:
                ids = tuple(
                    (set(ids) - set(exceptions['exclude']))
                    | set(exceptions['include'])
                )

            tz_ids_to_zips[ids].append(zip.zip)

        zips_to_tz_ids = {
            zip: ids for ids, zips in tz_ids_to_zips.items() for zip in zips
        }

        _dump_json_data(US_ZIPS_TO_TZ_IDS_MAP_PATH, zips_to_tz_ids)

    def generate_tz_name_to_tz_id_map(self):
        """Generate the map of timezone names to time zone identifiers."""
        tz_name_ids = {}
        for id in _progressbar(pytz.common_timezones):
            tz = pytz.timezone(id)
            tz_names = self._get_latest_tz_names(tz)
            if id.startswith('Australia'):
                tz_names = [
                    AU_MAP.get(tz_name, tz_name) for tz_name in tz_names
                ]
            if id.startswith('Asia'):
                tz_names = [
                    ASIA_MAP.get(tz_name, tz_name) for tz_name in tz_names
                ]

            # Include the aliases in the map, too
            for name in tz_names:
                tz_names.extend(self.tz_name_aliases.get(name, []))

            for name in tz_names:
                if name in tz_name_ids and id not in tz_name_ids[name]:
                    tz_name_ids[name].append(id)
                elif name not in tz_name_ids:
                    tz_name_ids[name] = [id]

        _dump_json_data(TZ_NAME_TO_TZ_IDS_MAP_PATH, tz_name_ids)

    def generate_offset_to_tz_id_map(self):
        """Generate the map of UTC offsets to time zone identifiers."""
        offset_tz_ids = defaultdict(list)
        for id in _progressbar(pytz.common_timezones):
            tz = pytz.timezone(id)
            offsets = self._get_latest_offsets(tz)
            for offset in offsets:
                offset_tz_ids[offset].append(id)

        _dump_json_data(OFFSET_TO_TZ_IDS_MAP_PATH, offset_tz_ids)


def _progressbar(lst):
    length_of_list = len(lst)
    for cnt, elem in enumerate(lst):
        yield elem
        if cnt % 10 == 9:
            stdout.write('\r%d/%d' % (cnt + 1, length_of_list))
            stdout.flush()


def _dump_json_data(path, data):
    with open(path, 'w') as f:
        json.dump(data, f, indent=2, sort_keys=True, separators=(',', ': '))



from setuptools import setup

with open("README.md", encoding="utf-8") as file:
    long_description = file.read()

setup(
    name='tz-trout',
    version='1.0.4.dev',
    url='http://github.com/closeio/tz-trout',
    license='MIT',
    author='Close.io',
    author_email='engineering@close.io',
    maintainer='Close.io',
    maintainer_email='engineering@close.io',
    description='Helps you figure out the time zone based on an address or a phone number.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    platforms='any',
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    packages=['tztrout'],
    package_data={'tztrout': ['data/*']},
    python_requires='>=3.8',
    install_requires=['phonenumbers>=8.3.0', 'python-dateutil', 'pytz'],
    extras_require={'dev': ['timezonefinder']},
    tests_require=['mock', 'pytest'],
)



import datetime
from unittest.mock import patch

import pytest

import tztrout

current_alaska_tz_name = (
    'AKDT' if tztrout.tz_ids_for_tz_name('AKDT') else 'AKST'
)
US_CA_TZ_NAMES = ['PT', 'MT', 'CT', 'ET', 'AT', current_alaska_tz_name]
AU_TZ_NAMES = ['AWT', 'ACT', 'AET']


def assert_only_one_tz(ids, expected_tz_name, tz_names):
    """Assert that a set of timezone identifiers exists in only one expected timezone across
    a given list of timezones.

    assert_only_one_tz(['America/New_York', America/Detroit'], 'ET', ['CT', 'ET', 'PT'])
    would pass because both 'America/New_York' and 'America/Detroit' are in the 'ET' timezone.

    assert_only_one_tz(['America/New_York', America/Los_Angeles'], 'ET', ['CT', 'ET', 'PT'])
    would fail because 'America/New_York' is in the ET timezone and 'America/Los_Angeles' is
    in the 'PT' timezone.
    """
    assert expected_tz_name in tz_names
    expected_tz_ids = set(tztrout.tz_ids_for_tz_name(expected_tz_name))

    # The given set of timezone ids should have a least one timezone id in commmon
    # with the expected timezone's timezone ids.
    ids = set(ids)
    assert ids & expected_tz_ids

    # The given set of timezone ids should not have any timezone ids in common with
    # any other (non-expected) timezones in the given timezone list.
    non_expected_tz_names = [i for i in tz_names if i != expected_tz_name]
    non_expected_tz_ids = set()
    for tz_name in non_expected_tz_names:
        tz_ids_for_name = set(tztrout.tz_ids_for_tz_name(tz_name))
        non_expected_tz_ids.update(tz_ids_for_name)

    assert not ids & non_expected_tz_ids


class FakeDateTime(datetime.datetime):
    """A datetime replacement that lets you set utcnow()"""

    @classmethod
    def utcnow(cls, *args, **kwargs):
        if hasattr(cls, 'dt'):
            return cls.dt
        raise NotImplementedError(
            'use FakeDateTime.set_utcnow(datetime) first'
        )

    @classmethod
    def set_utcnow(cls, dt):
        cls.dt = dt


class TestTZIdsForPhone:
    @pytest.mark.parametrize(
        ('phone', 'tz_ids'),
        [
            ('+1 (650) 333 4444', ['America/Los_Angeles']),
            ('+48 601 941 311', ['Europe/Warsaw']),
        ],
    )
    def test_ids_for_phone(self, phone, tz_ids):
        ids = tztrout.tz_ids_for_phone(phone)
        assert ids == tz_ids

    @pytest.mark.parametrize(
        ('phone', 'tz_name', 'tz_id'),
        [
            # United States -- Special cases to make sure ET is not counted as part of state timezone
            # Wisconsin
            ('+14143334444', 'CT', 'America/Chicago'),
            # Texas
            ('+12143334444', 'CT', 'America/Chicago'),
            # United States
            # New York, NY
            ('+12123334444', 'ET', 'America/New_York'),
            ('+16463334444', 'ET', 'America/New_York'),
            # Los Angeles, CA
            ('+18183334444', 'PT', 'America/Los_Angeles'),
            # Chicago, IL
            ('+16303334444', 'CT', 'America/Chicago'),
            # Houston, TX
            ('+17133334444', 'CT', 'America/Chicago'),
            # Philadelphia, PA
            ('+12153334444', 'ET', 'America/New_York'),
            # Phoenix, AZ
            ('+16023334444', 'MT', 'America/Phoenix'),
            # San Antonio, TX
            ('+12103334444', 'CT', 'America/Chicago'),
            # San Diego, CA
            ('+16193334444', 'PT', 'America/Los_Angeles'),
            # Dallas, TX
            ('+12143334444', 'CT', 'America/Chicago'),
            # San Jose, CA
            ('+14083334444', 'PT', 'America/Los_Angeles'),
            # Austin, TX
            ('+15123334444', 'CT', 'America/Chicago'),
            # Indianapolis, IN
            ('+13173334444', 'ET', 'America/Indiana/Indianapolis'),
            # Jacksonville, FL
            ('+19043334444', 'ET', 'America/New_York'),
            # San Francisco, CA
            ('+14153334444', 'PT', 'America/Los_Angeles'),
            # Columbus, OH
            ('+16143334444', 'ET', 'America/New_York'),
            # Charlotte, NC
            ('+17043334444', 'ET', 'America/New_York'),
            # Fort Worth, TX
            ('+16823334444', 'CT', 'America/Chicago'),
            # Detroit, MI
            ('+13133334444', 'ET', 'America/Detroit'),
            # El Paso, TX
            ('+19153334444', 'MT', 'America/Denver'),
            # Memphis, TN
            ('+19013334444', 'CT', 'America/Chicago'),
            # Denver, CO
            ('+13033334444', 'MT', 'America/Denver'),
            # Washington, DC
            ('+12023334444', 'ET', 'America/New_York'),
            # Canada
            # Toronto, ON
            ('+14163334444', 'ET', 'America/Toronto'),
            # Montreal, QC
            ('+15143334444', 'ET', 'America/Toronto'),
            # Calgary, AB
            ('+14033334444', 'MT', 'America/Edmonton'),
            # Ottawa, ON
            ('+13433334444', 'ET', 'America/Toronto'),
            ('+16133334444', 'ET', 'America/Toronto'),
            # Edmonton, AB
            ('+17803334444', 'MT', 'America/Edmonton'),
            # Mississauga, ON
            ('+12893334444', 'ET', 'America/Toronto'),
            # Winnipeg, MB
            ('+14313334444', 'CT', 'America/Winnipeg'),
            # Vancouver, BC
            ('+16043334444', 'PT', 'America/Vancouver'),
            # Halifax, NS
            ('+19023334444', 'AT', 'America/Halifax'),
            # Saskatoon, SK
            ('+13063334444', 'CT', 'America/Regina'),
        ],
    )
    def test_major_cities_us_ca(
        self,
        phone,
        tz_name,
        tz_id,
    ):
        """Make sure all the major cities in the United States and Canada match the right time zone
        (and the right time zone only).
        """
        ids = tztrout.tz_ids_for_phone(phone)
        assert_only_one_tz(ids, tz_name, US_CA_TZ_NAMES)
        assert set(ids) == {tz_id}

    @pytest.mark.parametrize(
        ('phone', 'tz_name'),
        [
            # NSW - New South Wales
            # WA - Western Australia
            # NT - Northern Territory
            # SA - South Australia
            # TAS - Tasmania
            # VIC - Victoria
            # ACT - Australian Capital Territory
            # QLD - Queensland
            # Sydney, NSW
            ('+61 27 333 4444', 'AET'),
            ('+61 28 333 4444', 'AET'),
            ('+61 29 333 4444', 'AET'),
            # Perth, WA
            ('+61 852 22 4444', 'AWT'),
            ('+61 853 22 4444', 'AWT'),
            ('+61 854 22 4444', 'AWT'),
            ('+61 861 22 4444', 'AWT'),
            ('+61 862 22 4444', 'AWT'),
            ('+61 863 22 4444', 'AWT'),
            ('+61 864 22 4444', 'AWT'),
            ('+61 865 22 4444', 'AWT'),
            # Darwin, NT
            ('+61 879 22 4444', 'ACT'),
            ('+61 889 22 4444', 'ACT'),
            # Adelaide, SA
            ('+61 870 22 4444', 'ACT'),
            ('+61 871 22 4444', 'ACT'),
            ('+61 872 22 4444', 'ACT'),
            ('+61 873 22 4444', 'ACT'),
            ('+61 874 22 4444', 'ACT'),
            ('+61 881 22 4444', 'ACT'),
            ('+61 882 22 4444', 'ACT'),
            ('+61 883 22 4444', 'ACT'),
            ('+61 884 22 4444', 'ACT'),
            # Hobart, TAS
            ('+61 361 22 4444', 'AET'),
            ('+61 362 22 4444', 'AET'),
            # Melbourne, VIC
            ('+61 37 333 4444', 'AET'),
            ('+61 38 333 4444', 'AET'),
            ('+61 39 333 4444', 'AET'),
            # Canberra, ACT
            ('+61 251 22 4444', 'AET'),
            ('+61 252 22 4444', 'AET'),
            ('+61 261 22 4444', 'AET'),
            ('+61 262 22 4444', 'AET'),
            # Brisbane, QLD
            ('+61 72 333 4444', 'AET'),
            ('+61 73 333 4444', 'AET'),
            # Townsville, QLD
            ('+61 744 22 4444', 'AET'),
            ('+61 745 22 4444', 'AET'),
            ('+61 777 22 4444', 'AET'),
        ],
    )
    def test_major_cities_australia(self, phone, tz_name):
        """Make sure all the major cities in Australia match the right time zone
        (and the right time zone only).
        """
        ids = tztrout.tz_ids_for_phone(phone)
        assert_only_one_tz(ids, tz_name, AU_TZ_NAMES)


class TestTZIdsForAddress:
    @pytest.mark.parametrize(
        (
            'country',
            'state',
            'city',
            'zipcode',
            'expected_tz_ids',
            'is_exact_match',
        ),
        [
            ('US', 'California', None, None, ['America/Los_Angeles'], True),
            ('US', 'CA', None, None, ['America/Los_Angeles'], True),
            ('US', 'CA', '', None, ['America/Los_Angeles'], True),
            ('US', 'CA', 'Palo Alto', None, ['America/Los_Angeles'], True),
            ('US', '', 'Palo Alto', None, ['America/Los_Angeles'], True),
            ('US', None, 'Palo Alto', None, ['America/Los_Angeles'], True),
            (
                'US',
                '',
                '',
                None,
                ['America/Los_Angeles', 'America/New_York'],
                False,
            ),
            (
                'US',
                None,
                None,
                None,
                ['America/Los_Angeles', 'America/New_York'],
                False,
            ),
            ('PL', None, None, None, ['Europe/Warsaw'], True),
            # Invalid state, assume any US tz
            (
                'US',
                'XX',
                None,
                None,
                ['America/Los_Angeles', 'America/New_York'],
                False,
            ),
            (
                'US',
                'XX',
                '',
                None,
                ['America/Los_Angeles', 'America/New_York'],
                False,
            ),
            # Invalid city with state, ignore city
            ('US', 'CA', 'XX', None, ['America/Los_Angeles'], True),
            # Invalid city without state, assume any US tz
            (
                'US',
                '',
                'XX',
                None,
                ['America/Los_Angeles', 'America/New_York'],
                False,
            ),
            (
                'US',
                None,
                'XX',
                None,
                ['America/Los_Angeles', 'America/New_York'],
                False,
            ),
            (
                'US',
                'California',
                None,
                '94041',
                ['America/Los_Angeles'],
                True,
            ),
            ('US', None, None, '94041', ['America/Los_Angeles'], True),
            ('US', None, None, 94041, ['America/Los_Angeles'], True),
            ('US', None, None, '94041-1191', ['America/Los_Angeles'], True),
            ('US', None, None, '0000', [], True),
            ('US', 'NY', None, '10065', ['America/New_York'], True),
            ('US', 'IL', None, '60642', ['America/Chicago'], True),
            ('US', 'DC', None, '20022', ['America/New_York'], True),
        ],
    )
    def test_ids_for_address(
        self, country, state, city, zipcode, expected_tz_ids, is_exact_match
    ):
        ids = tztrout.tz_ids_for_address(
            country, state=state, city=city, zipcode=zipcode
        )
        if is_exact_match:
            assert expected_tz_ids == ids
        else:
            for tz_id in expected_tz_ids:
                assert tz_id in ids

    @pytest.mark.parametrize(
        ('country', 'state', 'city', 'tz_name'),
        [
            # United States -- Special cases to make sure ET is not counted as part of state timezone
            ('US', 'WI', None, 'CT'),
            ('US', 'TX', None, 'CT'),
            # United States
            ('US', 'NY', 'New York', 'ET'),
            ('US', 'CA', 'Los Angeles', 'PT'),
            ('US', 'IL', 'Chicago', 'CT'),
            ('US', 'TX', 'Houston', 'CT'),
            ('US', 'PA', 'Philadelphia', 'ET'),
            ('US', 'AZ', 'Phoenix', 'MT'),
            ('US', 'TX', 'San Antonio', 'CT'),
            ('US', 'CA', 'San Diego', 'PT'),
            ('US', 'TX', 'Dallas', 'CT'),
            ('US', 'CA', 'San Jose', 'PT'),
            ('US', 'TX', 'Austin', 'CT'),
            ('US', 'IN', 'Indianapolis', 'ET'),
            ('US', 'FL', 'Jacksonville', 'ET'),
            ('US', 'CA', 'San Francisco', 'PT'),
            ('US', 'OH', 'Columbus', 'ET'),
            ('US', 'NC', 'Charlotte', 'ET'),
            ('US', 'TX', 'Fort Worth', 'CT'),
            ('US', 'MI', 'Detroit', 'ET'),
            ('US', 'TX', 'El Paso', 'MT'),
            ('US', 'TN', 'Memphis', 'CT'),
            ('US', 'CO', 'Denver', 'MT'),
            ('US', 'DC', 'Washington', 'ET'),
            # Canada
            ('CA', 'ON', 'Toronto', 'ET'),
            ('CA', 'QC', 'Montreal', 'ET'),
            ('CA', 'AB', 'Calgary', 'MT'),
            ('CA', 'ON', 'Ottawa', 'ET'),
            ('CA', 'AB', 'Edmonton', 'MT'),
            ('CA', 'ON', 'Mississauga', 'ET'),
            ('CA', 'MB', 'Winnipeg', 'CT'),
            ('CA', 'BC', 'Vancouver', 'PT'),
            ('CA', 'NS', 'Halifax', 'AT'),
            ('CA', 'SK', 'Saskatoon', 'CT'),
        ],
    )
    def test_major_cities_us_ca(
        self,
        country,
        state,
        city,
        tz_name,
    ):
        """Make sure all the major cities in the United States and Canada match the right time zone
        (and the right time zone only).
        """
        ids = tztrout.tz_ids_for_address(country, state=state, city=city)
        assert_only_one_tz(ids, tz_name, US_CA_TZ_NAMES)

    @pytest.mark.parametrize(
        ('state', 'city', 'tz_name'),
        [
            # NSW - New South Wales
            # WA - Western Australia
            # NT - Northern Territory
            # SA - South Australia
            # TAS - Tasmania
            # VIC - Victoria
            # ACT - Australian Capital Territory
            # QLD - Queensland
            ('NSW', 'Sydney', 'AET'),
            ('WA', 'Perth', 'AWT'),
            ('NT', 'Darwin', 'ACT'),
            ('SA', 'Adelaide', 'ACT'),
            ('TAS', 'Hobart', 'AET'),
            ('VIC', 'Melbourne', 'AET'),
            ('ACT', 'Canberra', 'AET'),
            ('QLD', 'Brisbane', 'AET'),
            ('QLD', 'Townsville', 'AET'),
        ],
    )
    def test_major_cities_australia(self, state, city, tz_name):
        """Make sure all the major cities in Australia match the right time zone
        (and the right time zone only).
        """
        ids = tztrout.tz_ids_for_address('AU', state=state, city=city)
        assert_only_one_tz(ids, tz_name, AU_TZ_NAMES)

    def test_australia_without_state_info(self):
        ids = tztrout.tz_ids_for_address('AU')
        expected_ids = [
            "Australia/Sydney",
            "Australia/Perth",
            "Australia/Darwin",
            "Australia/Adelaide",
            "Australia/Darwin",
            "Australia/Adelaide",
            "Australia/Hobart",
            "Australia/Melbourne",
            "Australia/Sydney",
            "Australia/Brisbane",
        ]
        for tz_id in expected_ids:
            assert tz_id in ids

    def test_canada_without_state_info(self):
        ids = tztrout.tz_ids_for_address('CA')
        expected_ids = [
            "America/Whitehorse",
            "America/Vancouver",
            "America/Edmonton",
            "America/Regina",
            "America/Winnipeg",
            "America/Iqaluit",
            "America/Toronto",  # "America/Montreal", TODO re-add Montreal when pytz.country_timezones is fixed
            "America/Moncton",
            "America/Halifax",
            "America/St_Johns",
        ]
        for tz_id in expected_ids:
            assert tz_id in ids

    @pytest.mark.parametrize(
        ('zipcode', 'expected_tz_name'),
        [
            ('00501', 'ET'),
            ('01434', 'ET'),
            ('01473', 'ET'),
            ('02120', 'ET'),
            ('04046', 'ET'),
            ('05201', 'ET'),
            ('06108', 'ET'),
            ('07677', 'ET'),
            ('08501', 'ET'),
            ('10065', 'ET'),
            ('10709', 'ET'),
            ('11542', 'ET'),
            ('12019', 'ET'),
            ('13052', 'ET'),
            ('14741', 'ET'),
            ('15552', 'ET'),
            ('16701', 'ET'),
            ('17250', 'ET'),
            ('18769', 'ET'),
            ('19399', 'ET'),
            ('20022', 'ET'),
            ('20153', 'ET'),
            ('20598', 'ET'),
            ('21619', 'ET'),
            ('22485', 'ET'),
            ('23509', 'ET'),
            ('24147', 'ET'),
            ('25503', 'ET'),
            ('26568', 'ET'),
            ('27619', 'ET'),
            ('28409', 'ET'),
            ('29351', 'ET'),
            ('30334', 'ET'),
            ('31764', 'ET'),
            ('32224', 'ET'),
            ('33758', 'ET'),
            ('34280', 'ET'),
            ('35218', 'CT'),
            ('36177', 'CT'),
            ('37304', 'ET'),
            ('38569', 'CT'),
            ('39553', 'CT'),
            ('40964', 'ET'),
            ('41543', 'ET'),
            ('42086', 'CT'),
            ('43338', 'ET'),
            ('44799', 'ET'),
            ('45448', 'ET'),
            ('46135', 'ET'),
            ('47250', 'ET'),
            ('48919', 'ET'),
            ('50039', 'CT'),
            ('51555', 'CT'),
            ('52101', 'CT'),
            ('53561', 'CT'),
            ('54546', 'CT'),
            ('55128', 'CT'),
            ('56473', 'CT'),
            ('57661', 'MT'),
            ('58079', 'CT'),
            ('59038', 'MT'),
            ('60586', 'CT'),
            ('60642', 'CT'),
            ('61477', 'CT'),
            ('62089', 'CT'),
            ('63024', 'CT'),
            ('64674', 'CT'),
            ('65264', 'CT'),
            ('66505', 'CT'),
            ('67578', 'CT'),
            ('68348', 'CT'),
            ('69170', 'CT'),
            ('70175', 'CT'),
            ('71772', 'CT'),
            ('72629', 'CT'),
            ('73622', 'CT'),
            ('74051', 'CT'),
            ('75287', 'CT'),
            ('76164', 'CT'),
            ('77380', 'CT'),
            ('78799', 'CT'),
            ('78851', 'CT'),
            ('81132', 'MT'),
            ('82443', 'MT'),
            ('83637', 'MT'),
            ('84761', 'MT'),
            ('85671', 'MT'),
            ('86440', 'MT'),
            ('87901', 'MT'),
            ('88023', 'MT'),
            ('89496', 'PT'),
            ('90310', 'PT'),
            ('91950', 'PT'),
            ('92262', 'PT'),
            ('93552', 'PT'),
            ('94080', 'PT'),
            ('95968', 'PT'),
            ('96079', 'PT'),
            ('97470', 'PT'),
            ('98109', 'PT'),
            ('98194', 'PT'),
            ('99950', current_alaska_tz_name),
        ],
    )
    def test_tz_names_for_zipcode_only_addresses(
        self, zipcode, expected_tz_name
    ):
        ids = tztrout.tz_ids_for_address('US', zipcode=zipcode)
        assert_only_one_tz(ids, expected_tz_name, US_CA_TZ_NAMES)


class TestTZIdsForTZName:
    @pytest.mark.parametrize('tz_name', ['PT', 'PACIFIC'])
    def test_ids_for_tz_name(self, tz_name):
        pacific_ids = [
            'America/Dawson',
            'America/Fort_Nelson',
            'America/Los_Angeles',
            'America/Metlakatla',
            'America/Tijuana',
            'America/Vancouver',
            'America/Whitehorse',
            'Canada/Pacific',
            'US/Pacific',
        ]
        ids = tztrout.tz_ids_for_tz_name(tz_name)
        assert ids == pacific_ids


class TestLocalTimeForAddress:
    @patch('datetime.datetime', FakeDateTime)
    def test_local_time_in_spain(self):
        """Make sure local time is properly calculated for Spain."""
        FakeDateTime.set_utcnow(
            datetime.datetime(2016, 9, 13, 22, 15)
        )  # 15:15 PT / 22:15 UTC / 00:15 CEST
        local_time = tztrout.local_time_for_address('ES', city='Barcelona')
        assert str(local_time) == '2016-09-14 00:15:00+02:00'


class TestOffsetRangesForLocalTime:
    @patch('datetime.datetime', FakeDateTime)
    @pytest.mark.parametrize(
        (
            'hour_now',
            'local_time_start',
            'local_time_end',
            'expected_offset_results',
        ),
        [
            (
                20,
                datetime.time(9),
                datetime.time(17),
                [[-11 * 60, -3 * 60], [13 * 60, 14 * 60]],
            ),
            (
                0,
                datetime.time(9),
                datetime.time(17),
                [[9 * 60, 14 * 60], [-14 * 60, -7 * 60]],
            ),
            (
                4,
                datetime.time(9),
                datetime.time(17),
                [[5 * 60, 13 * 60], [-14 * 60, -11 * 60]],
            ),
            (7, datetime.time(9), datetime.time(17), [[2 * 60, 10 * 60]]),
            (
                20,
                datetime.time(17),
                datetime.time(9),
                [[-3 * 60, 13 * 60], [-14 * 60, -11 * 60]],
            ),
            (0, datetime.time(17), datetime.time(9), [[-7 * 60, 9 * 60]]),
            (
                4,
                datetime.time(17),
                datetime.time(9),
                [[13 * 60, 14 * 60], [-11 * 60, 5 * 60]],
            ),
            (
                7,
                datetime.time(17),
                datetime.time(9),
                [[10 * 60, 14 * 60], [-14 * 60, 2 * 60]],
            ),
            (20, '9am', '5pm', [[-11 * 60, -3 * 60], [13 * 60, 14 * 60]]),
            (20, '9:00', '17:00', [[-11 * 60, -3 * 60], [13 * 60, 14 * 60]]),
        ],
    )
    def test_offset_ranges(
        self,
        hour_now,
        local_time_start,
        local_time_end,
        expected_offset_results,
    ):
        FakeDateTime.set_utcnow(datetime.datetime(2013, 1, 1, hour_now))
        offset_ranges = tztrout.offset_ranges_for_local_time(
            local_time_start, local_time_end
        )
        assert offset_ranges == expected_offset_results


class TestNonDSTOffsetsForPhone:
    @pytest.mark.parametrize(
        ('phone', 'result'),
        [('+1 650 333 4444', [-8 * 60]), ('+1 212 333 4444', [-5 * 60])],
    )
    def test_non_dst_offsets_for_phone(self, phone, result):
        offsets = tztrout.non_dst_offsets_for_phone(phone)
        assert offsets == result


class TestNonDSTOffsetsForAddress:
    @pytest.mark.parametrize(
        ('state', 'result'), [('CA', [-8 * 60]), ('NY', [-5 * 60])]
    )
    def test_non_dst_offsets_for_address(self, state, result):
        offsets = tztrout.non_dst_offsets_for_address('US', state=state)
        assert offsets == result
